import xml.etree.ElementTree as ET

class WavIXMLFormat:
    """
    iXML recorder metadata, as defined by iXML 2.0
    """
    def __init__(self, xml):
        self.source = xml
        self.parsed = ET.fromstring(xml)

    def _get_text_value(self, xpath):
        e = self.parsed.find("./" + xpath)
        if e is not None:
            return e.text
    
    @property
    def project(self):
        return self._get_text_value("PROJECT")
        
    @property
    def scene(self):
        return self._get_text_value("SCENE")

    @property
    def take(self):
        return self._get_text_value("TAKE")

    @property
    def tape(self):
        return self._get_text_value("TAPE")

    @property
    def family_uid(self):
        return self._get_text_value("FILE_SET/FAMILY_UID")

    @property
    def family_name(self):
        return self._get_text_value("FILE_SET/FAMILY_NAME")






import struct

from collections import namedtuple

class ListChunkDescriptor(namedtuple('ListChunkDescriptor' , 'signature children')):

    def find(chunk_path):
        if len(chunk_path) > 1:
            for chunk in self.children:
                if type(chunk) is ListChunkDescriptor and \
                        chunk.signature is chunk_path[0]:
                            return chunk.find(chunk_path[1:])
        else:
            for chunk in self.children:
                if type(chunk) is ChunkDescriptor and \
                        chunk.ident is chunk_path[0]:
                            return chunk


class ChunkDescriptor(namedtuple('ChunkDescriptor', 'ident start length') ):
    def read_data(self, from_stream):
        from_stream.seek(self.start)
        return from_stream.read(self.length)

def parse_list_chunk(stream, length):
    start = stream.tell()

    signature = stream.read(4)

    children = []
    while (stream.tell() - start) < length:
        children.append(parse_chunk(stream))

    return ListChunkDescriptor(signature=signature, children=children)

def parse_chunk(stream):
    ident = stream.read(4)
    if len(ident) != 4:
        return

    sizeb = stream.read(4)
    size  = struct.unpack('<I',sizeb)[0]

    displacement = size
    if displacement % 2 is not 0:
        displacement = displacement + 1

    if ident in [b'RIFF',b'LIST']:
        return parse_list_chunk(stream=stream, length=size)
    else:
        start = stream.tell()
        stream.seek(displacement,1)
        return ChunkDescriptor(ident=ident, start=start, length=size)








import struct

from collections import namedtuple

from .wave_parser import parse_chunk, ChunkDescriptor, ListChunkDescriptor
from .wave_ixml_reader import WavIXMLFormat
from .wave_bext_reader import WavBextReader
from .wave_info_reader import WavInfoChunkReader

WavDataDescriptor = namedtuple('WavDataDescriptor','byte_count frame_count')

WavInfoFormat = namedtuple("WavInfoFormat",'audio_format channel_count sample_rate byte_rate block_align bits_per_sample')

WavBextFormat = namedtuple("WavBextFormat",'description originator originator_ref ' +
    'originator_date originator_time time_reference version umid ' +
    'loudness_value loudness_range max_true_peak max_momentary_loudness max_shortterm_loudness ' +
    'coding_history')


class WavInfoReader():
    """
    format : WAV format
    bext   : The Broadcast-WAV extension as definied by EBU Tech 3285 v2 (2011)

    """

    def __init__(self, path, info_encoding='latin_1', bext_encoding='ascii'):
        """
        Parse a WAV audio file for metadata.

        * `path`: A filesystem path to the wav file you wish to probe.

        * `info_encoding`: The text encoding of the INFO metadata fields.
          `latin_1`/Win CP1252 has always been a pretty good guess for this.

        * `bext_encoding`: The text encoding to use when decoding the string
          fields of the Broadcast-WAV extension. Per EBU 3285 this is ASCII
          but this parameter is available to you if you encounter a werido.

        """
        with open(path, 'rb') as f:
            chunks = parse_chunk(f)

            self.main_list = chunks.children
            f.seek(0)

            self.fmt    = self._get_format(f)
            self.bext   = self._get_bext(f, encoding=bext_encoding)
            self.ixml   = self._get_ixml(f)
            self.info   = self._get_info(f, encoding=info_encoding)
            self.data   = self._describe_data(f)

    def _find_chunk_data(self, ident, from_stream, default_none=False):
        chunk_descriptor = None
        top_chunks = (chunk for chunk in self.main_list if type(chunk) is ChunkDescriptor)

        if default_none:
            chunk_descriptor = next((chunk for chunk in top_chunks if chunk.ident == ident),None)
        else:
            chunk_descriptor = next((chunk for chunk in top_chunks if chunk.ident == ident))

        if chunk_descriptor:
            return chunk_descriptor.read_data(from_stream)
        else:
            return None


    def _describe_data(self,f):
        data_chunk = next(c for c in self.main_list if c.ident == b'data')

        return WavDataDescriptor(byte_count= data_chunk.length,
                frame_count= int(data_chunk.length / self.fmt.block_align))


    def _get_format(self,f):
        fmt_data = self._find_chunk_data(b'fmt ',f)

        # The format chunk is
        # audio_format    U16
        # channel_count   U16
        # sample_rate     U32   Note an integer
        # byte_rate       U32   == SampleRate * NumChannels * BitsPerSample/8
        # block_align     U16   == NumChannels * BitsPerSample/8
        # bits_per_sampl  U16
        packstring = "<HHIIHH"
        rest_starts = struct.calcsize(packstring)

        unpacked = struct.unpack(packstring, fmt_data[:rest_starts])

        #0x0001	WAVE_FORMAT_PCM	PCM
        #0x0003	WAVE_FORMAT_IEEE_FLOAT	IEEE float
        #0x0006	WAVE_FORMAT_ALAW	8-bit ITU-T G.711 A-law
        #0x0007	WAVE_FORMAT_MULAW	8-bit ITU-T G.711 µ-law
        #0xFFFE	WAVE_FORMAT_EXTENSIBLE	Determined by SubFormat

        #https://sno.phy.queensu.ca/~phil/exiftool/TagNames/RIFF.html
        return WavInfoFormat(audio_format = unpacked[0],
                    channel_count   = unpacked[1],
                    sample_rate     = unpacked[2],
                    byte_rate       = unpacked[3],
                    block_align     = unpacked[4],
                    bits_per_sample = unpacked[5]
                    )

    def _get_info(self, f, encoding):
        finder = (chunk.signature for chunk in self.main_list \
                if type(chunk) is ListChunkDescriptor)

        if b'INFO' in finder:
            return WavInfoChunkReader(f, encoding)

    def _get_bext(self, f, encoding):
        bext_data = self._find_chunk_data(b'bext',f,default_none=True)
        return WavBextReader(bext_data, encoding)

    def _get_ixml(self,f):
        ixml_data = self._find_chunk_data(b'iXML',f,default_none=True)
        if ixml_data is None:
            return None

        ixml_string = ixml_data.decode('utf-8')
        return WavIXMLFormat(ixml_string)








import struct

class WavBextReader:
    def __init__(self,bext_data,encoding):
        # description[256]

        # originator[32]
        # originatorref[32]
        # originatordate[10]   "YYYY:MM:DD"
        # originatortime[8]    "HH:MM:SS"
        # lowtimeref U32
        # hightimeref U32
        # version U16
        # umid[64]
        #
        # EBU 3285 fields
        # loudnessvalue S16    (in LUFS*100)
        # loudnessrange S16    (in LUFS*100)
        # maxtruepeak   S16    (in dbTB*100)
        # maxmomentaryloudness S16 (LUFS*100)
        # maxshorttermloudness S16 (LUFS*100)
        # reserved[180]
        # codinghistory []
        if bext_data is None:
            return None

        packstring = "<256s"+ "32s" + "32s" + "10s" + "8s" + "QH" + "64s" + "hhhhh" + "180s"

        rest_starts = struct.calcsize(packstring)
        unpacked = struct.unpack(packstring, bext_data[:rest_starts])

        def sanatize_bytes(bytes):
            first_null = next( (index for index, byte in enumerate(bytes) if byte == 0 ), None )
            if first_null is not None:
                trimmed = bytes[:first_null]
            else:
                trimmed = bytes

            decoded = trimmed.decode(encoding)
            return decoded

        bext_version = unpacked[6]
        if bext_version > 0:
            self.umid = unpacked[6]
        else:
            self.umid = None

        if bext_version > 1:
            self.loudness_value          = unpacked[8] / 100.0,
            self.loudness_range          = unpacked[9] / 100.0
            self.max_true_peak           = unpacked[10] / 100.0
            self.max_momentary_loudness  = unpacked[11] / 100.0
            self.max_shortterm_loudness  = unpacked[12] / 100.0
        else:
            self.loudness_value          = None
            self.loudness_range          = None
            self.max_true_peak           = None
            self.max_momentary_loudness  = None
            self.max_shortterm_loudness  = None

            self.description     = sanatize_bytes(unpacked[0])
            self.originator      = sanatize_bytes(unpacked[1])
            self.originator_ref  = sanatize_bytes(unpacked[2])
            self.originator_date = sanatize_bytes(unpacked[3])
            self.originator_time = sanatize_bytes(unpacked[4])
            self.time_reference  = unpacked[5]
            self.version         = unpacked[6]
            self.coding_history  = sanatize_bytes(bext_data[rest_starts:])


    def to_dict(self):
        return {'description':      self.description,
                'originator':       self.originator,
                'originator_ref':   self.originator_ref,
                'originator_date':  self.originator_date,
                'originator_time':  self.originator_time,
                'time_reference':   self.time_reference,
                'version':          self.version,
                'coding_history':   self.coding_history,
                'loudness_value':   self.loudness_value,
                'loudness_range':   self.loudness_range,
                'max_true_peak':    self.max_true_peak,
                'max_momentary_loudness':   self.max_momentary_loudness,
                'max_shortterm_loudness':   self.max_shortterm_loudness
                }

